/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2012                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/
/*! \file trimesh_base.cpp
\ingroup code_sample

\brief the minimal example of using the lib

This file contain a minimal example of the library

*/

#include<vcg/complex/complex.h>

// input output
#include<wrap/io_trimesh/import_ply.h>
#include<wrap/io_trimesh/export_ply.h>

// topology computation
#include<vcg/complex/algorithms/update/topology.h>
#include<vcg/complex/algorithms/update/normal.h>
#include<vcg/complex/algorithms/update/quality.h>
#include<vcg/complex/algorithms/update/color.h>

using namespace vcg;
using namespace std;

class MyEdge;
class MyFace;
class MyVertex;
struct MyUsedTypes : public UsedTypes<	Use<MyVertex>   ::AsVertexType,
                                        Use<MyEdge>     ::AsEdgeType,
                                        Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public Vertex<MyUsedTypes, vertex::Coord3f, vertex::Normal3f,
    vertex::Qualityf, vertex::Color4b, vertex::BitFlags  >{};
class MyFace    : public Face< MyUsedTypes, face::FFAdj,  face::VertexRef, face::BitFlags > {};
class MyEdge    : public Edge<MyUsedTypes>{};
class MyMesh    : public tri::TriMesh< vector<MyVertex>, vector<MyFace> , vector<MyEdge>  > {};

int main( int argc, char **argv )
{
  if(argc<2)
  {
    printf("Usage trimesh_base <meshfilename.obj>\n");
    return -1;
  }

  MyMesh m;

  if(tri::io::ImporterPLY<MyMesh>::Open(m,argv[1])!=0)
  {
    printf("Error reading file  %s\n",argv[1]);
    exit(0);
  }

  tri::UpdateNormal<MyMesh>::PerVertexNormalized(m);
  tri::UpdateTopology<MyMesh>::FaceFace(m);

  printf("Input mesh  vn:%i fn:%i\n",m.VN(),m.FN());
  printf( "Mesh has %i vert and %i faces\n", m.VN(), m.FN() );

  // First Check: No non-manifold edge must exists!
  int nonManifCnt=0;
  for(MyMesh::FaceIterator fi = m.face.begin();fi!=m.face.end();++fi)
  {
    for(int i=0;i<3;++i)
    {
      if((*fi).FFp(i) == &*fi) continue;
      MyFace *fp =  (*fi).FFp(i);
      int faceInd = (*fi).FFi(i);
      if(fp->FFp(faceInd) != &*fi) // Non Manifold!
        nonManifCnt++;
    }
  }
  if(nonManifCnt>0)
  {
      printf("Mesh has non manifold edges!\n");
      exit(-1);
  }

  // First Step: Count number of incident faces on each vertex
  tri::UpdateQuality<MyMesh>::VertexConstant(m,0);
  for(MyMesh::FaceIterator fi = m.face.begin();fi!=m.face.end();++fi)
  {
    for(int i=0;i<3;++i)
       (*fi).V(i)->Q()++;
  }

  // Just to debug: save a mesh with per vertex color mapping the degree of vertices
  tri::UpdateColor<MyMesh>::PerVertexQualityRamp(m);
  tri::io::ExporterPLY<MyMesh>::Save(m,"out.ply",tri::io::Mask::IOM_VERTCOLOR);

  // Main checking loop.
  // For each vertex check the consistency of incident face number
  // Use a PerVertex visiting bit to avoid multiple counting of non manifold vertices

  int nonManifVertCnt=0;
  tri::UpdateColor<MyMesh>::PerVertexConstant(m,Color4b::LightGray);

  // clear the per vertex visiting bit for all the vertices
  tri::UpdateFlags<MyMesh>::VertexClearV(m);

  for(MyMesh::FaceIterator fi = m.face.begin();fi!=m.face.end();++fi)
  {
    for(int i=0;i<3;++i)
    {
      if(!(*fi).V(i)->IsV())
      {
        (*fi).V(i)->SetV();
        // Now loop around a vertex V using FlipF/FlipE to count the number
        // of faces incident on V that can be reached using just FF adjancency
        face::Pos<MyFace> startPos(&*fi,i,(*fi).V(i));
        face::Pos<MyFace> iterPos = startPos;
        int faceCnt=0;
        bool borderFlag=false;
        do
        {
          if(iterPos.IsBorder()) borderFlag = true;
          iterPos.FlipF();
          iterPos.FlipE();
          faceCnt++;
        }
        while(iterPos != startPos);

        if(borderFlag)
        {
          assert((faceCnt%2) ==0);
          faceCnt/=2;
        }

        if(faceCnt!=(*fi).V(i)->Q())
        {
          (*fi).V(i)->C()=Color4b::Red; // just to debug Color non manif vertex in red
          nonManifVertCnt++;
        }
      }
    }
  }
  printf("Mesh has %i non manifold vertexes\n",nonManifVertCnt);
  tri::io::ExporterPLY<MyMesh>::Save(m,"out2.ply",tri::io::Mask::IOM_VERTCOLOR);

  return 0;
}
